//******************************************************************************
//
// File:    Chat.java
// Package: edu.rit.chat2
// Unit:    Class edu.rit.chat2.Chat
//
// This Java source file is copyright (C) 2006 by Alan Kaminsky. All rights
// reserved. For further information, contact the author, Alan Kaminsky, at
// ark@cs.rit.edu.
//
// This Java source file is part of the M2MI Library ("The Library"). The
// Library is free software; you can redistribute it and/or modify it under the
// terms of the GNU General Public License as published by the Free Software
// Foundation; either version 2 of the License, or (at your option) any later
// version.
//
// The Library is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
// details.
//
// A copy of the GNU General Public License is provided in the file gpl.txt. You
// may also obtain a copy of the GNU General Public License on the World Wide
// Web at http://www.gnu.org/licenses/gpl.html or by writing to the Free
// Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
// USA.
//
//******************************************************************************

package edu.rit.chat2;

import edu.rit.m2mi.InvocationException;
import edu.rit.m2mi.M2MI;

import edu.rit.util.Timer;
import edu.rit.util.TimerTask;
import edu.rit.util.TimerThread;
import edu.rit.util.Transcript;

import java.io.IOException;
import java.io.PrintStream;

import java.util.ArrayList;
import java.util.Random;

import javax.swing.JOptionPane;

/**
 * Class Chat is a rudimentary M2MI-based chat application. The program displays
 * a simple chat UI (an instance of class {@linkplain ChatFrame}), and the
 * program exports a chat object (an instance of class Chat) that implements
 * interface {@linkplain ChatRef}. Each chat room is associated with its own
 * multihandle for interface ChatRef, and the chat object is attached to the
 * multihandle corresponding to the chat room the user joined. When the user
 * sends a line of text in the UI, the line is broadcast to all the chat objects
 * in the chat room by calling <TT>putLine()</TT> on the multihandle. When each
 * chat object receives a <TT>putLine()</TT> invocation, it displays the line of
 * text in the chat log in its UI.
 * <P>
 * To discover which chat rooms are out there, the chat object also implements
 * interface {@linkplain ChatDiscoveryRef}. Each chat object periodically
 * broadcasts the name and multihandle of its chat room by calling
 * <TT>reportChatRoom()</TT> on an omnihandle for interface ChatDiscoveryRef.
 * When each chat object receives a <TT>reportChatRoom()</TT> invocation, it
 * puts the chat room name and multihandle in a list. When the user selects a
 * chat room, the chat object detaches itself from the previous chat room's
 * multihandle and attaches itself to the new chat room's multihandle.
 * <P>
 * The chat demo is intended merely to demonstrate M2MI omnihandle and
 * multihandle invocations and is not intended to be a full-featured chat
 * application.
 * <P>
 * Usage: java edu.rit.chat2.Chat <I>username</I>
 * <P>
 * When running the chat demo application, M2MI and M2MP must be configured, and
 * the M2MP Daemon must be running if necessary. See packages {@linkplain
 * edu.rit.m2mi} and {@linkplain edu.rit.m2mp} for further information.
 * <P>
 * The chat demo application displays this UI:
 * <P>
 * <CENTER>
 * <IMG SRC="doc-files/chatui.png">
 * </CENTER>
 * <P>
 * Pick the chat room to join from the drop-down list, or click the "New..."
 * button to create a new chat room. Type a line of text into any chat window on
 * any machine and hit return or click the "Send" button. The line of text,
 * prefixed by the user name, shows up in all the chat windows on all the
 * machines that have joined the same chat room.
 *
 * @author  Alan Kaminsky
 * @version 27-Mar-2006
 */
public class Chat
	implements ChatFrameListener, ChatRef, ChatDiscoveryRef, TimerTask
	{

// Hidden constants.

	// Range of times for a quick broadcast of a chat room's existence.
	private static final int QUICK_MIN_TIME = 500;
	private static final int QUICK_RANGE    = 500;

	// Range of times for a normal broadcast of a chat room's existence.
	private static final int NORMAL_MIN_TIME = 10000;
	private static final int NORMAL_RANGE    = 10000;

// Hidden data members.

	// For UI.
	private String myUserName;
	private ChatFrame myChatFrame;
	private Transcript myChatLog;
	private PrintStream myChatLogStream;

	// List of chat rooms out there, in the order they appear in the UI list.
	private ArrayList<ChatRoom> myChatRoomList = new ArrayList<ChatRoom>();

	// Chat room the user has joined, or null if not joined.
	private ChatRoom myChatRoom;

	// For broadcasting the existence of myChatRoom.
	private Timer myTimer = TimerThread.getDefault().createTimer (this);
	private Random myPrng = new Random (System.currentTimeMillis());
	private ChatDiscoveryRef allChatDiscovery;

// Hidden constructors.

	private Chat()
		{
		}

// Exported operations.

	/**
	 * Run this chat application.
	 *
	 * @param  args  Command line arguments.
	 */
	public void run
		(String[] args)
		throws Throwable
		{
		// Parse command line arguments.
		if (args.length != 1) usage();
		myUserName = args[0];

		// Initialize the M2MI Layer.
		M2MI.initialize();

		// Set up UI.
		myChatFrame = new ChatFrame ("M2MI Chat Demo 2 -- " + args[0]);
		myChatLog = myChatFrame.getChatLog();
		myChatLogStream = new PrintStream (myChatLog.getOutputStream());
		myChatFrame.setListener (this);

		// Export this chat object so it can receive M2MI invocations for
		// interface ChatDiscoveryRef.
		M2MI.export (this, ChatDiscoveryRef.class);

		// Set up an omnihandle for broadcasting M2MI invocations for interface
		// ChatDiscoveryRef.
		allChatDiscovery = (ChatDiscoveryRef)
			M2MI.getOmnihandle (ChatDiscoveryRef.class);

		// Discover existing chat rooms.
		allChatDiscovery.requestChatRooms();
		}

	/**
	 * Report that the user wants to create a new chat room.
	 */
	public synchronized void newChatRoom()
		{
		// Prompt user for chat room name.
		String name =
			JOptionPane.showInputDialog
				(myChatFrame,
				 "Chat room name:",
				 "New Chat Room",
				 JOptionPane.QUESTION_MESSAGE);

		if (name != null && ! name.equals (""))
			{
			// Set up new chat room object with name and a new multihandle.
			ChatRoom chatroom =
				new ChatRoom
					(name,
					 (ChatRef) M2MI.getMultihandle (ChatRef.class));

			// Add chat room to UI list and select it.
			myChatRoomList.add (chatroom);
			myChatFrame.addChatRoom (chatroom, true);
			}
		}

	/**
	 * Report that the user selected the given chat room.
	 *
	 * @param  i  Chat room index.
	 */
	public synchronized void setChatRoom
		(int i)
		{
		// Do nothing if no chat room was selected.
		if (i < 0) return;

		// Do nothing if the current chat room was selected.
		ChatRoom chatroom = myChatRoomList.get (i);
		if (chatroom == myChatRoom) return;

		// Leave the previous chat room if any.
		if (myChatRoom != null) myChatRoom.disjoin (this);

		// Join the new chat room.
		myChatRoom = chatroom;
		myChatRoom.join (this);

		// Clear previous contents of chat log.
		try { myChatLog.clear(); } catch (IOException exc) {}

		// Restart timer for broadcasting chat room's existence.
		beginQuickTimeout();
		}

	/**
	 * Process the given line of text entered by the user.
	 *
	 * @param  line  Line of text.
	 */
	public synchronized void send
		(String line)
		{
		try
			{
			myChatRoom.putMessage (myUserName + "> " + line);
			}
		catch (InvocationException exc)
			{
			exc.printStackTrace (System.err);
			}
		}

	/**
	 * Display the given line of text in this chat object's chat log.
	 *
	 * @param  line  Line of text.
	 */
	public synchronized void putMessage
		(String line)
		{
		myChatLogStream.println (line);
		}

	/**
	 * Request everyone to report the existence of their chat rooms.
	 */
	public synchronized void requestChatRooms()
		{
		if (myChatRoom != null) beginQuickTimeout();
		}

	/**
	 * Report that a chat room exists.
	 *
	 * @param  theChatRoom  Chat room (name plus multihandle).
	 */
	public synchronized void reportChatRoom
		(ChatRoom theChatRoom)
		{
		// Do nothing if we already know about the chat room.
		if (myChatRoomList.contains (theChatRoom)) return;

		// Add chat room to UI list but don't select it.
		myChatRoomList.add (theChatRoom);
		myChatFrame.addChatRoom (theChatRoom, false);

		// Restart broadcast timer if it's the current chat room.
		if (theChatRoom.equals (myChatRoom)) beginNormalTimeout();
		}

	/**
	 * Upon timeout, broadcast the existence of this chat room.
	 *
	 * @param  theTimer  Timer.
	 */
	public synchronized void action
		(Timer theTimer)
		{
		if (myTimer.isTriggered())
			{
			allChatDiscovery.reportChatRoom (myChatRoom);
			beginNormalTimeout();
			}
		}

// Hidden operations.

	/**
	 * Begin a normal timeout for broadcasting the current chat room's
	 * existence.
	 */
	private void beginNormalTimeout()
		{
		myTimer.start (NORMAL_MIN_TIME + myPrng.nextInt (NORMAL_RANGE));
		}

	/**
	 * Begin a quick timeout for broadcasting the current chat room's existence.
	 */
	private void beginQuickTimeout()
		{
		myTimer.start (QUICK_MIN_TIME + myPrng.nextInt (QUICK_RANGE));
		}

	/**
	 * Print a usage message and exit.
	 */
	private void usage()
		{
		System.err.println ("Usage: java edu.rit.chat2.Chat <username>");
		System.exit (1);
		}

// Main program.

	/**
	 * Main program.
	 *
	 * @param  args  Command line arguments.
	 */
	public static void main
		(String[] args)
		throws Throwable
		{
		new Chat().run (args);
		}

	}
