//******************************************************************************
//
// File:    Survey.java
// Package: edu.rit.survey
// Unit:    Class edu.rit.survey.Survey
//
// This Java source file is copyright (C) 2006 by Alan Kaminsky. All rights
// reserved. For further information, contact the author, Alan Kaminsky, at
// ark@cs.rit.edu.
//
// This Java source file is part of the M2MI Library ("The Library"). The
// Library is free software; you can redistribute it and/or modify it under the
// terms of the GNU General Public License as published by the Free Software
// Foundation; either version 2 of the License, or (at your option) any later
// version.
//
// The Library is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
// details.
//
// A copy of the GNU General Public License is provided in the file gpl.txt. You
// may also obtain a copy of the GNU General Public License on the World Wide
// Web at http://www.gnu.org/licenses/gpl.html or by writing to the Free
// Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
// USA.
//
//******************************************************************************

package edu.rit.survey;

import edu.rit.m2mi.Eoid;
import edu.rit.m2mi.M2MI;

import edu.rit.util.Timer;
import edu.rit.util.TimerTask;
import edu.rit.util.TimerThread;

import java.util.HashMap;
import java.util.LinkedList;
import java.util.Random;

import javax.swing.JFrame;

/**
 * Class Survey is the Online Survey application, an ad hoc collaborative
 * application using M2MI.
 * <P>
 * Usage: java edu.rit.survey.Survey
 * <P>
 * When running the Online Survey application, M2MI and M2MP must be configured,
 * and the M2MP Daemon must be running if necessary. See packages {@linkplain
 * edu.rit.m2mi} and {@linkplain edu.rit.m2mp} for further information.
 * <P>
 * The Online Survey application displays this UI:
 * <P>
 * <CENTER>
 * <IMG SRC="doc-files/surveyui.png">
 * </CENTER>
 * <P>
 * The drop-down list has all the surveys that are out there. Pick one from the
 * list, or click the "New..." button to create a new survey. The survey
 * question and answers are displayed below, along with the number of votes for
 * each answer. To vote for a certain answer, click the corresponding button.
 *
 * @author  Alan Kaminsky
 * @version 25-Mar-2006
 */
public class Survey
	implements SurveyRef, SurveyUIListener
	{

// Hidden constants.

	// Range of times for a quick broadcast of a survey state.
	private static final int QUICK_MIN_TIME = 500;
	private static final int QUICK_RANGE    = 500;

	// Range of times for a normal broadcast of a survey state.
	private static final int NORMAL_MIN_TIME = 10000;
	private static final int NORMAL_RANGE    = 10000;

// Hidden data members.

	// Reference to this survey object, for use as a synchronization lock.
	private Survey mySelf;

	// Survey UI.
	private SurveyUI myUI;

	// List of survey info objects in the order they appear in the UI.
	private LinkedList<SurveyInfo> mySurveyInfoList =
		new LinkedList<SurveyInfo>();

	// Mapping from survey state object to survey info object.
	private HashMap<SurveyState,SurveyInfo> mySurveyInfoMap =
		new HashMap<SurveyState,SurveyInfo>();

	// Currently displayed survey info.
	private SurveyInfo mySurveyInfo;

	// Omnihandle for reporting survey states.
	private SurveyRef allSurveys;

	// For timed broadcasts of survey states.
	private TimerThread myTimerThread;
	private Random myPrng;

	// New survey dialog.
	private NewSurveyDialog myNewSurveyDialog;

// Hidden helper classes.

	/**
	 * Class SurveyInfo is a record of information associated with a survey.
	 * This includes the complete state of the survey, the participant's ID
	 * within the survey, and a timer for survey state broadcasts.
	 *
	 * @author  Alan Kaminsky
	 * @version 25-Mar-2006
	 */
	private class SurveyInfo
		implements TimerTask
		{
		private SurveyState mySurveyState;
		private Eoid myID;
		private Timer myTimer;

		/**
		 * Create a new survey info object containing the given survey state
		 * object.
		 *
		 * @param  theSurveyState  Survey state object.
		 */
		public SurveyInfo
			(SurveyState theSurveyState)
			{
			// Set up fields.
			mySurveyState = theSurveyState;
			myID = Eoid.next();
			mySurveyState.setSurveyVote (myID, new SurveyVote());
			myTimer = myTimerThread.createTimer (this);

			// Record new info object.
			mySurveyInfoList.add (this);
			mySurveyInfoMap.put (mySurveyState, this);
			myUI.addSurvey (mySurveyState);

			// Begin normal timeout for broadcasting survey state.
			beginNormalTimeout();
			}

		/**
		 * Merge just-reported survey state into existing survey state.
		 *
		 * @param  theSurveyState  Just-reported survey state.
		 */
		public void merge
			(SurveyState theSurveyState)
			{
			// Merge survey states.
			mySurveyState.merge (theSurveyState);

			// Update the UI if it's displaying this survey state.
			if (this == mySurveyInfo)
				{
				displayVotes();
				}

			// Begin normal timeout for broadcasting survey state.
			beginNormalTimeout();
			}

		/**
		 * Report that the user voted for the given answer.
		 *
		 * @param  i  Answer index (0-3).
		 */
		public void setVote
			(int i)
			{
			// Update survey state.
			SurveyVote vote = mySurveyState.getSurveyVote (myID);
			vote.setVoteFor (i);
			mySurveyState.recalculate();
			myUI.setVoteCounts (mySurveyState);
			myUI.setVotes (vote);

			// Do a quick broadcast of the new survey state.
			beginQuickTimeout();
			}

		/**
		 * Display this survey info object's survey state in the UI.
		 */
		public void displaySurvey()
			{
			myUI.setQuestionAnswers (mySurveyState);
			displayVotes();
			}

		/**
		 * Display this survey info object's survey votes in the UI.
		 */
		private void displayVotes()
			{
			myUI.setVoteCounts (mySurveyState);
			myUI.setVotes (mySurveyState.getSurveyVote (myID));
			}

		/**
		 * Begin a normal timeout for broadcasting this survey info object's
		 * survey state.
		 */
		private void beginNormalTimeout()
			{
			myTimer.start (NORMAL_MIN_TIME + myPrng.nextInt (NORMAL_RANGE));
			}

		/**
		 * Begin a quick timeout for broadcasting this survey info object's
		 * survey state.
		 */
		private void beginQuickTimeout()
			{
			myTimer.start (QUICK_MIN_TIME + myPrng.nextInt (QUICK_RANGE));
			}

		/**
		 * Upon timeout, broadcast this survey info object's survey state. Also
		 * begin a normal timeout for the next broadcast.
		 */
		public void action
			(Timer theTimer)
			{
			synchronized (mySelf)
				{
				if (myTimer.isTriggered())
					{
					allSurveys.reportSurvey (mySurveyState);
					beginNormalTimeout();
					}
				}
			}
		}

// Hidden constructors.

	/**
	 * Construct a new Online Survey application.
	 */
	private Survey()
		{
		}

// Exported operations.

	/**
	 * Run this Online Survey application.
	 */
	public synchronized void run
		(String[] args)
		throws Throwable
		{
		// Set up synchronization lock.
		mySelf = this;

		// Initialize M2MI.
		M2MI.initialize();

		// Set up UI.
		myUI = new SurveyUI();
		myUI.setListener (this);
		myUI.setDefaultCloseOperation (JFrame.EXIT_ON_CLOSE);
		myUI.setVisible (true);

		// Set up omnihandle for reporting survey states.
		allSurveys = (SurveyRef) M2MI.getOmnihandle (SurveyRef.class);

		// Export this object so it can receive omnihandle invocations.
		M2MI.export (this, SurveyRef.class);

		// Set up for timed broadcasts of survey states.
		myTimerThread = TimerThread.getDefault();
		myPrng = new Random (System.currentTimeMillis());

		// Set up new survey dialog.
		myNewSurveyDialog = new NewSurveyDialog (myUI);
		}

	/**
	 * Report the state of a survey.
	 *
	 * @param  theSurveyState  Survey state.
	 */
	public synchronized void reportSurvey
		(SurveyState theSurveyState)
		{
		// Get survey info.
		SurveyInfo info = mySurveyInfoMap.get (theSurveyState);

		if (info == null)
			{
			// This is a new survey. Set up new survey info.
			info = new SurveyInfo (theSurveyState);
			}

		else
			{
			// This is an existing survey. Merge just-reported survey state into
			// existing survey state.
			info.merge (theSurveyState);
			}
		}

	/**
	 * Report that the user wants to create a new survey.
	 */
	public void newSurvey()
		{
		SurveyState surveystate = myNewSurveyDialog.getSurveyState();
		if (surveystate != null)
			{
			allSurveys.reportSurvey (surveystate);
			}
		}

	/**
	 * Report that the user selected the given survey.
	 *
	 * @param  i  Survey index.
	 */
	public synchronized void setSurvey
		(int i)
		{
		mySurveyInfo = mySurveyInfoList.get (i);
		mySurveyInfo.displaySurvey();
		}

	/**
	 * Report that the user voted for the given answer.
	 *
	 * @param  i  Answer index (0-3).
	 */
	public synchronized void setVote
		(int i)
		{
		mySurveyInfo.setVote (i);
		}

// Hidden operations.

	/**
	 * Print a usage message and exit.
	 */
	private void usage()
		{
		System.err.println ("Usage: java edu.rit.survey.Survey [<name> <question> <answer1> <answer2> <answer3> <answer4>]");
		System.exit (1);
		}

// Main program.

	/**
	 * Main program.
	 */
	public static void main
		(String[] args)
		throws Throwable
		{
		new Survey().run (args);
		}

	}
